/* 
   nuvfile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
   Mercilessly ripped from nuvrec (roman@mars.tuwien.ac.at)
   by Justin Schoeman (c) 2001

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <dlfcn.h>
#include <time.h>
#include "nuppelvideo.h"
#include "minilzo.h"
#include "RTjpegN.h"


#ifdef HAVE_CONVERT
#include "../convert.h"
#endif
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#include "../nvrec.h"

#define KEYFRAMEDIST 30

#define HEAP_ALLOC(var,size) \
	long __LZO_MMODEL var [ ((size) + (sizeof(long) - 1)) / sizeof(long) ]
#define IN_LEN (1024*1024L)
#define OUT_LEN (IN_LEN + IN_LEN / 64 + 16 + 3)

struct nuvfile_core_t {
	int fd;

	unsigned long long tpf;
	int effectivedsp;
	unsigned long long atime;
	unsigned long long vtime;
	int frameofgop;
	
	struct rtframeheader frameH;

	char *basename;
	unsigned int split_size;
	double frame_rate;
	int vwidth;
	int vheight;
	int qfactor;
	
	int M1;
	int M2;
	unsigned char * cbuf;
	HEAP_ALLOC(wrkmem,LZO1X_1_MEM_COMPRESS);
	lzo_byte out [ OUT_LEN ];

#ifdef HAVE_CONVERT
	convert_t * conv;
#endif	
};

struct file_core_t nuvfile_core;

int nuvfile_int_create(char *fname, int number, int w, int h, int Q, double fps)
{
	struct rtfileheader fileheader;
	struct rtframeheader frameheader;
	char realfname[255];
	static unsigned long int tbls[128];
	static const char finfo[12] = "NuppelVideo";
	static const char vers[5]   = "0.05";
	int ostr;
	
	memset(&fileheader, 0, sizeof(fileheader));
	memset(&frameheader, 0, sizeof(frameheader));
	if (number == 0) { 
		RTjpeg_init_compress(tbls, w, h, Q);
		RTjpeg_init_mcompress();
		snprintf(realfname, 250, "%s", fname);
	} else {
		snprintf(realfname, 250, "%s-%d", fname, number);
	}
	ostr = open(realfname, O_RDWR);
	if(ostr == -1) {
		ostr = open(realfname, O_RDWR|O_CREAT|O_TRUNC, 0666);
		if (ostr == -1) {
			PERROR("Could not open file");
			return -1;
		}
	}
	
	memcpy(fileheader.finfo, finfo, sizeof(fileheader.finfo));
	memcpy(fileheader.version, vers, sizeof(fileheader.version));
	fileheader.width = w;
	fileheader.height = h;
	fileheader.desiredwidth = 0;
	fileheader.desiredheight = 0;
	fileheader.pimode = 'P';
	fileheader.aspect = 1.0;
	fileheader.fps = fps;
	fileheader.videoblocks = -1;
	fileheader.audioblocks = -1;
	fileheader.textsblocks = 0;
	fileheader.keyframedist = KEYFRAMEDIST;
	write(ostr, &fileheader, FILEHEADERSIZE);
	
	frameheader.frametype = 'D';
	frameheader.comptype = 'R';
	frameheader.packetlength = sizeof(tbls);
	write(ostr, &frameheader, FRAMEHEADERSIZE);
	write(ostr, tbls, sizeof(tbls));
	fsync(ostr);
	
	return ostr;
}

int nuvfile_int_keyframe(struct nuvfile_core_t * nuvc)
{
	int is_key = 0;
	
	nuvc->frameH.keyframe = nuvc->frameofgop;
	if(nuvc->frameofgop % KEYFRAMEDIST == 0) {
		
		is_key = 1;
		
		write(nuvc->fd, "RTjjjjjjjjjjjjjjjjjjjjjjjj", FRAMEHEADERSIZE);

		nuvc->frameofgop = 0;
		nuvc->frameH.keyframe = 0;

		nuvc->frameH.frametype = 'S';
		nuvc->frameH.comptype = 'V';
		nuvc->frameH.filters = 0;
		nuvc->frameH.packetlength = 0;
		nuvc->frameH.timecode = nuvc->vtime / 1000000ULL;
		write(nuvc->fd, &nuvc->frameH, FRAMEHEADERSIZE);

		nuvc->frameH.frametype = 'S';
		nuvc->frameH.comptype = 'A';
		nuvc->frameH.filters = 0;
		nuvc->frameH.packetlength = 0;
		nuvc->frameH.timecode = nuvc->effectivedsp;
		write(nuvc->fd, &nuvc->frameH, FRAMEHEADERSIZE);
	}
	nuvc->frameofgop++;
	nuvc->frameH.timecode = nuvc->vtime / 1000000ULL;
	nuvc->vtime += nuvc->tpf;
	return is_key;
	
}

void nuvfile_core_close(struct file_core_t * fct)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	off_t flength;
	if(!nuvc) return;
	if(nuvc->cbuf)
		free(nuvc->cbuf);
	flength = lseek(nuvc->fd, 0, SEEK_CUR);
	if(flength != -1) {
		ftruncate(nuvc->fd, flength);
	} else {
		EPRINT("Could not determine file length - file may be corrupt.\n");
	}
	close(nuvc->fd);
	
#ifdef HAVE_CONVERT
	if(nuvc->conv) {
		convert_end(nuvc->conv);
	}
#endif	
	
	free(nuvc);
	free(fct);
}

void nuvfile_core_nullvideo(struct file_core_t * fct)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	
	if(!nuvc)
		return;

	nuvfile_int_keyframe(nuvc);

	nuvc->frameH.frametype = 'V';
	nuvc->frameH.comptype  = 'L';
	nuvc->frameH.packetlength = 0;
	write(nuvc->fd, &nuvc->frameH, FRAMEHEADERSIZE);
}


void nuvfile_core_rotate(struct file_core_t * fct)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	
	if(!nuvc)
		return;

	char extension[16];
	char *name_buffer;
	struct tm *ptr;
	time_t tm;
	int len;
	
	tm = time(NULL);
	ptr = gmtime(&tm);
	strftime(extension, 21, "%Y%m%dT%H%M%S.avi", ptr);
	len = strlen(nuvc->basename) + 1;
	
	name_buffer = (char *)malloc(len*sizeof(char)+21);
	snprintf(name_buffer, len+21, "%s_%s", nuvc->basename, extension);

	close(nuvc->fd);

	nuvc->fd = nuvfile_int_create(name_buffer, 0, nuvc->vwidth, nuvc->vheight, nuvc->qfactor, nuvc->frame_rate);
	
	if(!nuvc->fd) PERROR("Could not open file");
	/* FIXME: Need to free the nuvc structure */
	free(name_buffer);
	
	return;
}

void nuvfile_core_fastvideo(struct file_core_t * fct, unsigned char **buf)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	int i;
	int is_key = 0;
	
	if(!nuvc)
		return;

#ifdef HAVE_CONVERT
	if(nuvc->conv) {
		buf = convert_convert(nuvc->conv, buf);
		if(!buf) {
			EPRINT("Error converting image!\n");
			return;
		}
	}
#endif

	is_key = nuvfile_int_keyframe(nuvc);
	
	if (is_key)
		if(nuvc->basename && (fct->file_core_size(fct) > nuvc->split_size))
			nuvfile_core_rotate(fct);
	
	
	i = RTjpeg_mcompressYUV420(nuvc->cbuf, buf[0], nuvc->M1, nuvc->M2);
	nuvc->frameH.frametype = 'V';
	nuvc->frameH.filters = 0;
	nuvc->frameH.comptype  = '1';
	nuvc->frameH.packetlength = i;
	write(nuvc->fd, &nuvc->frameH, FRAMEHEADERSIZE);
	write(nuvc->fd, nuvc->cbuf, i);
}

void nuvfile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	int i, j, k;
	int is_key = 0;
	
	if(!nuvc) return;
	
#ifdef HAVE_CONVERT
	if(nuvc->conv) {
		buf = convert_convert(nuvc->conv, buf);
		if(!buf) {
			EPRINT("Error converting image!\n");
			return;
		}
	}
#endif
	is_key = nuvfile_int_keyframe(nuvc);
	
	if (is_key)
		if(nuvc->basename && (fct->file_core_size(fct) > nuvc->split_size))
			nuvfile_core_rotate(fct);
	
	i = RTjpeg_mcompressYUV420(nuvc->cbuf, buf[0], nuvc->M1, nuvc->M2);
	j = lzo1x_1_compress(nuvc->cbuf, i, nuvc->out, &k, nuvc->wrkmem);
	nuvc->frameH.frametype = 'V';
	nuvc->frameH.filters = 0;
	nuvc->frameH.comptype  = '2';
	nuvc->frameH.packetlength = k;
	write(nuvc->fd, &nuvc->frameH, FRAMEHEADERSIZE);
	write(nuvc->fd, nuvc->out, k);
}

void nuvfile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	
	if(!nuvc)
		return;

	nuvc->frameH.timecode = nuvc->atime / 1000000ULL;
	nuvc->atime += nuvc->tpf;

	nuvc->frameH.frametype = 'A';
	nuvc->frameH.comptype = '0';
	nuvc->frameH.packetlength = n;

	write(nuvc->fd, &nuvc->frameH, FRAMEHEADERSIZE);
	write(nuvc->fd, buf, n);
}

void nuvfile_core_sync(struct file_core_t * fct)
{
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	fsync(nuvc->fd);
}


static unsigned int nuvfile_core_formats[] = {NV_YUV420, NV_YUV422, NV_RGB24, NV_END_FORMAT};
//static unsigned int nuvfile_core_formats[] = {NV_YUV422, NV_RGB24, NV_END_FORMAT};

unsigned int *nuvfile_core_format(char **args)
{
/* Parse the command line to find the format that will be needed */
	return nuvfile_core_formats;
}

long nuvfile_core_size(struct file_core_t * fct)
{
	/* Get the file size */
	struct nuvfile_core_t * nuvc = (struct nuvfile_core_t *)fct->priv;
	struct stat buf;
	
	if (fstat(nuvc->fd, &buf) < 0){
		PERROR("fstat() failed over nuvfile open descriptor");
		return 0;
	}
	return (long)buf.st_size;
		
}


void nuvfile_core_usage()
{
	MPRINT("NUV file core options:\n");
	MPRINT("\t-vg <key frame rate>  [25]\n");
	MPRINT("\t-vq <quality factor>  [100]\n");
	MPRINT("\t-vl <luma factor>     [2]\n");
	MPRINT("\t-vc <chroma factor>   [1]\n");
	MPRINT("\t-t split output size (in Mbytes)            [600]\n");
#ifdef HAVE_CONVERT
	MPRINT("\t-vhook \"module name and params\"             [NONE]\n");
#endif
	MPRINT("\t-o <output filename>                        [NONE]\n");
	MPRINT("\t * filename starting with split://base indicates base name\n");

}

struct file_core_t * nuvfile_core_init(char ** args, unsigned int format, ntime_t vrate)
{
	struct file_core_t * fct;
	struct nuvfile_core_t * nuvc;
	int arate = -1;
	int abits = -1;
	int achans = -1;
	char *fname = NULL;
	int len = 0;
	int keyrate = 25;
	char *vhook = NULL;
	unsigned int split = NV_ROTATE_NUPPEL_SIZE;

	NV_ARGS;

	fct =  malloc(sizeof(struct file_core_t));
	if(!fct) {
		PERROR("Could no malloc new cap core");
		return NULL;
	}
	nuvc = malloc(sizeof(struct nuvfile_core_t));
	if(!nuvc) {
		PERROR("Could no malloc new nuv core");
		free(fct);
		return NULL;
	}

	memcpy(fct, &nuvfile_core, sizeof(struct file_core_t));
	fct->priv = (void *)nuvc;
	memset(nuvc, 0, sizeof(struct nuvfile_core_t));
	
	nuvc->vwidth = -1;
	nuvc->vheight = -1;
	nuvc->qfactor = 255;

	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -r.\n");
				goto usage;
			}
			arate = atoi(NV_THIS_ARG);
		}else if(strcmp(NV_THIS_ARG, "-t") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -t.\n");
				goto usage;
			}
			split = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -b.\n");
				goto usage;
			}
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -w.\n");
				goto usage;
			}
			nuvc->vwidth = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -h.\n");
				goto usage;
			}
			nuvc->vheight = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-vg") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vg.\n");
				goto usage;
			}
			keyrate = atoi(NV_THIS_ARG);
			if(keyrate < 0) {
				EPRINT("Invalid argument to -vg (>=0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vq") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vq.\n");
				goto usage;
			}
			nuvc->qfactor = atoi(NV_THIS_ARG);
			nuvc->qfactor *= 5;
			nuvc->qfactor /= 2;
			nuvc->qfactor += 5;
			if((nuvc->qfactor < 16) || (nuvc->qfactor > 255)) {
				EPRINT("Invalid argument to -vq (10-100).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vl") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vl.\n");
				goto usage;
			}
			nuvc->M1 = atoi(NV_THIS_ARG);
			if(nuvc->M1 < 0) {
				EPRINT("Invalid argument to -vl (>=0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-vc") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vc.\n");
				goto usage;
			}
			nuvc->M2 = atoi(NV_THIS_ARG);
			if(nuvc->M2 < 0) {
				EPRINT("Invalid argument to -vc (>=0).\n");
				goto usage;
			}
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -o.\n");
				goto usage;
			}
			fname = NV_THIS_ARG;
		} else if(strcmp(NV_THIS_ARG, "-vr") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vr.\n");
				goto usage;
			}
			vrate = (unsigned long long)(1e9 / atof(NV_THIS_ARG));
		} else if(strcmp(NV_THIS_ARG, "-vhook") == 0) {
			if(!NV_NEXT_ARG) {
				EPRINT("Missing argument to -vhook.\n");
				goto usage;
			}
			vhook = NV_THIS_ARG;
		}
	}
	
	if(!fname) {
		EPRINT("Missing -o <filename> argument.\n");
		goto usage;
	}
	nuvc->basename = NULL;
	
	if(strncmp(fname, "split://", 8) == 0) {
		
		char extension[20];
		char *name_buffer;
		struct tm *ptr;
		time_t tm;
		
		nuvc->split_size = split*1048576;
		IPRINT("Split Size is %d\n", nuvc->split_size);
		len = strlen(fname+8) + 1;
		nuvc->basename = (char *)malloc(len*sizeof(char));
		strcpy(nuvc->basename, (fname+8));
		tm = time(NULL);
		ptr = gmtime(&tm);
		strftime(extension, 20, "%Y%m%dT%H%M%S.avi", ptr);
		name_buffer = (char *)malloc(len+21);
		snprintf(name_buffer, len+21, "%s_%s", nuvc->basename, extension);
		fname = name_buffer;
	}

	IPRINT("File: %s (%dx%d)\n", fname, nuvc->vwidth, nuvc->vheight);

	if(format != NV_YUV420) {
#ifdef HAVE_CONVERT
		IPRINT("No native format found - initialise format conversion\n");
		nuvc->conv = convert_new(format, NV_YUV420, nuvc->vwidth, nuvc->vheight, vhook);
		if(!nuvc->conv) {
			EPRINT("Could not init format conversion core!\n");
			goto error0;
		}
#else
		EPRINT("No native format found, and conversion not available - giving up.\n");
		goto error0;
#endif
	}

#ifdef HAVE_CONVERT
	if((!nuvc->conv) && (vhook)) {
		IPRINT("Instantiating conversion function for vhooks...\n");
		nuvc->conv = convert_new(format, format, nuvc->vwidth, nuvc->vheight, vhook);
		if(!nuvc->conv) {
			EPRINT("Could not init format conversion core!\n");
			goto error0;
		}
	}
#endif

	nuvc->frame_rate = 1e9 / (double)vrate;
	IPRINT("Frame rate is %f\n", nuvc->frame_rate);

	nuvc->fd = nuvfile_int_create(fname, 0, nuvc->vwidth, nuvc->vheight, nuvc->qfactor, nuvc->frame_rate);
	if(!nuvc->fd) {
		PERROR("Could not open file");
		goto error0;
	}
	
	nuvc->tpf = vrate;
	nuvc->effectivedsp = arate * 100;

	nuvc->cbuf = malloc(nuvc->vwidth * nuvc->vheight * 3);
	if(!nuvc->cbuf) {
		PERROR("Could not allocate compression buffer");
		goto error1;
	}
	
	if(lzo_init() != LZO_E_OK) {
		EPRINT("Could not init lzo - falling back to standard RTjpeg\n");
		fct->file_core_video = nuvfile_core_fastvideo;
	}
	
	return (void *)fct;

 error1:
	close(nuvc->fd);
 error0:
#ifdef HAVE_CONVERT
	if(nuvc->conv) {
		convert_end(nuvc->conv);
	}
#endif
	free(fct);
	free(nuvc);
	return NULL;

 usage:
	EPRINT("Error parsing options.  Please run without any command line options for\n");
	EPRINT("detailed usage information.\n");
	free(fct);
	free(nuvc);
	return NULL;
}


void nuvfile_core_overrides(char **args)
{
	nvcore_add_arg(args, "-b");
	nvcore_add_arg(args, "16");
	nvcore_add_arg(args, "-r");
	nvcore_add_arg(args, "44100");
	nvcore_add_arg(args, "-s");
	return;
}


struct file_core_t nuvfile_core = {
	file_core_init: nuvfile_core_init,
	file_core_close: nuvfile_core_close,
	file_core_nullvideo: nuvfile_core_nullvideo,
	file_core_fastvideo: nuvfile_core_fastvideo,
	file_core_video: nuvfile_core_video,
	file_core_audio: nuvfile_core_audio,
	file_core_sync: nuvfile_core_sync,
	file_core_format: nuvfile_core_format,
	file_core_overrides: nuvfile_core_overrides,
	file_core_size : nuvfile_core_size,
	file_core_usage: nuvfile_core_usage,
	priv: NULL
};
